/* Copyright (c) 2010-2015. The SimGrid Team.
 * All rights reserved.                                                     */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#include "simgrid/jedule/jedule_platform.hpp"
#include "simgrid/jedule/jedule_output.hpp"
#include "simgrid/s4u/As.hpp"
#include "simgrid/host.h"

#include "xbt/asserts.h"
#include "xbt/dynar.h"
#include "xbt/str.h"
#include <string>

#include <stdlib.h>
#include <stdio.h>

#if HAVE_JEDULE

static xbt_dict_t host2_simgrid_parent_container;
static xbt_dict_t container_name2container;

namespace simgrid {
namespace jedule {

Container::Container(std::string name)
: name(name) {

  xbt_dict_set(container_name2container, this->name.c_str(), this, nullptr);
}

void Container::addChild(jed_container_t child){
  xbt_assert(this != nullptr);
  xbt_assert(child != nullptr);
  this->children.push_back(child);
  child->parent = this;
}

void Container::addResources(std::vector<sg_host_t> hosts){
  this->is_lowest = 1;
  this->children.clear();
  this->last_id = 0;

  //FIXME do we need to sort?: xbt_dynar_sort_strings(host_names);

  for (auto host : hosts) {
    const char *host_name = sg_host_get_name(host);
    this->name2id.insert({host_name, this->last_id});
    (this->last_id)++;
    xbt_dict_set(host2_simgrid_parent_container, host_name, this, nullptr);
    this->resource_list.push_back(host);
  }
}

void Container::createHierarchy(AS_t from_as){
  xbt_dict_cursor_t cursor = nullptr;
  char *key;
  AS_t elem;
  xbt_dict_t routing_sons = from_as->children();

  if (xbt_dict_is_empty(routing_sons)) {
    // I am no AS
    // add hosts to jedule platform
    xbt_dynar_t table = from_as->hosts();
    unsigned int dynar_cursor;
    sg_host_t host;

    std::vector<sg_host_t> hosts;

    xbt_dynar_foreach(table, dynar_cursor, host) {
      hosts.push_back(host);
    }
    this->addResources(hosts);
    xbt_dynar_free(&table);
  } else {
    xbt_dict_foreach(routing_sons, cursor, key, elem) {
      jed_container_t child_container = new simgrid::jedule::Container(std::string(elem->name()));
      this->addChild(child_container);
      child_container->createHierarchy(elem);
    }
  }
}

std::vector<int> Container::getHierarchy() {
  xbt_assert( this!= nullptr );

  if(this->parent != nullptr ) {

    if(!this->parent->children.empty()) {
      // we are in the last level
      return this->parent->getHierarchy();
    } else {
      unsigned int i =0;
      int child_nb = -1;

      for (auto child : this->parent->children) {
        if( child == this) {
          child_nb = i;
          break;
        }
        i++;
      }

      xbt_assert( child_nb > - 1);
      std::vector<int> heir_list = this->parent->getHierarchy();
      heir_list.insert(heir_list.begin(), child_nb);
      return heir_list;
    }
  } else {
    int top_level = 0;
    std::vector<int> heir_list = {top_level};
    return heir_list;
  }
}

std::string Container::getHierarchyAsString(){
  std::string output("");

  std::vector<int> heir_list = this->getHierarchy();

  unsigned int length = heir_list.size();
  unsigned int i = 0;
  for (auto id : heir_list) {
    output += std::to_string(id);
    if( i != length-1 ) {
      output += ".";
    }
  }

  return output;
}

void Container::printResources(FILE * jed_file){
  unsigned int i=0;
  xbt_assert(!this->resource_list.empty());

  unsigned int res_nb = this->resource_list.size();
  std::string resid = this->getHierarchyAsString();

  fprintf(jed_file, "      <rset id=\"%s\" nb=\"%u\" names=\"", resid.c_str(), res_nb);
  for (auto res: this->resource_list) {
    const char * res_name = sg_host_get_name(res);
    fprintf(jed_file, "%s", res_name);
    if( i != res_nb-1 ) {
      fprintf(jed_file, "|");
    }
    i++;
  }
  fprintf(jed_file, "\" />\n");
}

void Container::print(FILE* jed_file) {
  xbt_assert( this != nullptr );
  fprintf(jed_file, "    <res name=\"%s\">\n", this->name.c_str());
  if( !this->children.empty()){
    for (auto child: this->children) {
      child->print(jed_file);
    }
  } else {
    this->printResources(jed_file);
  }
  fprintf(jed_file, "    </res>\n");
}

}
}

static int compare_ids(const void *num1, const void *num2) {
  return *((int*) num1) - *((int*) num2);
}

/* FIXME: That should be the destructor, shouldnt it? */
static void jed_free_container(jed_container_t container) {
  if(!container->children.empty()) {
    for (auto child: container->children){
      jed_free_container(child);
    }
  }
  delete container;
}

static void add_subset_to(xbt_dynar_t subset_list, int start, int end, jed_container_t parent) {
  jed_res_subset_t subset;

  xbt_assert( subset_list != nullptr );
  xbt_assert( parent != nullptr );

  subset = xbt_new0(s_jed_res_subset_t,1);
  subset->start_idx = start;
  subset->nres      = end-start+1;
  subset->parent    = parent;

  xbt_dynar_push(subset_list, &subset);
}

static void add_subsets_to(xbt_dynar_t subset_list, xbt_dynar_t hostgroup, jed_container_t parent) {
  unsigned int iter;
  char *host_name;
  int id;

  // get ids for each host
  // sort ids
  // compact ids
  // create subset for each id group

  xbt_assert( host2_simgrid_parent_container != nullptr );
  xbt_assert( subset_list != nullptr );
  xbt_assert( hostgroup != nullptr );
  xbt_assert( parent != nullptr );

  xbt_dynar_t id_list = xbt_dynar_new(sizeof(int), nullptr);

  xbt_dynar_foreach(hostgroup, iter, host_name) {
    jed_container_t parent;
    xbt_assert( host_name != nullptr );
    parent = (jed_container_t)xbt_dict_get(host2_simgrid_parent_container, host_name);
    id = parent->name2id.at(host_name);
    xbt_dynar_push(id_list, &id);
  }
  int nb_ids = xbt_dynar_length(id_list);
  xbt_dynar_sort(id_list, &compare_ids);

  int *id_ar = static_cast<int*>(xbt_dynar_to_array(id_list));

  if( nb_ids > 0 ) {
    int start = 0;

    int pos = start;
    for(int i=0; i<nb_ids; i++) {
      if( id_ar[i] - id_ar[pos] > 1 ) {
        add_subset_to( subset_list, id_ar[start], id_ar[pos], parent );
        start = i;

        if( i == nb_ids-1 ) {
          add_subset_to( subset_list, id_ar[i], id_ar[i], parent );
        }
      } else {
        if( i == nb_ids-1 ) {
          add_subset_to( subset_list, id_ar[start], id_ar[i], parent );
        }
      }
      pos = i;
    }
  }

  xbt_free(id_ar);
}

void jed_simgrid_get_resource_selection_by_hosts(xbt_dynar_t subset_list, std::vector<sg_host_t> *host_list) {
  std::unordered_map<const char*, xbt_dynar_t> parent2hostgroup;  // group hosts by parent

  xbt_assert( host_list != nullptr );

  // for each host name
  //  find parent container
  //  group by parent container

  for (auto host: *host_list) {
    const char *host_name = sg_host_get_name(host);
    jed_container_t parent = (jed_container_t)xbt_dict_get(host2_simgrid_parent_container, host_name);
    xbt_assert( parent != nullptr );

    auto host_group = parent2hostgroup.find(parent->name.c_str());
    if (host_group == parent2hostgroup.end()){
      xbt_dynar_t group = xbt_dynar_new(sizeof(char*), nullptr);
      xbt_dynar_push(group, &host_name);
      parent2hostgroup.insert({parent->name.c_str(), group});
    } else {
      xbt_dynar_push(host_group->second, &host_name);
    }
  }

  for (auto elm: parent2hostgroup) {
    jed_container_t parent = (jed_container_t)xbt_dict_get(container_name2container, elm.first);
    add_subsets_to(subset_list, elm.second, parent);
    xbt_dynar_free_container(&elm.second);
  }
}

void jedule_add_meta_info(jedule_t jedule, char *key, char *value) {
  xbt_assert(key != nullptr);
  xbt_assert(value != nullptr);

  jedule->jedule_meta_info.insert({key, value});
}

void jed_create_jedule(jedule_t *jedule) {
  *jedule = xbt_new0(s_jedule_t,1);
  host2_simgrid_parent_container = xbt_dict_new_homogeneous(nullptr);
  container_name2container       = xbt_dict_new_homogeneous(nullptr);
}

void jed_free_jedule(jedule_t jedule) {
  jed_free_container(jedule->root_container);

  xbt_free(jedule);

  xbt_dict_free(&host2_simgrid_parent_container);
  xbt_dict_free(&container_name2container);
}
#endif
